import numpy as np
import numba as nb


@nb.njit()
def calcCentroids(X, labels, n_clusters):
    """

    :param X:
    :param labels:
    :param n_clusters:
    :return:
    """
    centroids = np.zeros((labels.shape[0], n_clusters))
    n = n_clusters
    for i in range(n_clusters):
        mask = (labels == i)
        if not np.any(mask):
            n += -1
            centroids[i, :] = np.random.rand(n_clusters)
        else:
            clust = X[mask, :]
            cent = np.mean(clust, axis=0)
            centroids[i, :] = cent

    # if n != n_clusters:
    #     print('Some clusters unassigned')
    #     print('Assigned Clusters: ', n)

    return np.array(centroids)


@nb.njit()
def calcCosCentroids(X, labels, n_clusters):
    """

    :param X:
    :param labels:
    :param n_clusters:
    :return:
    """
    centroids = np.zeros((n_clusters, n_clusters))
    n = n_clusters
    for i in range(n_clusters):
        mask = (labels == i)
        if not np.any(mask):
            n += -1
            centroids[i, :] = np.random.rand(n_clusters)
        else:
            clust = X[mask, :]
            c = np.sum(clust, axis=0)
            cent = c / np.linalg.norm(c)
            centroids[i, :] = cent

    # if n != n_clusters:
    #     print('Some clusters unassigned')
    #     print('Assigned Clusters: ', n)

    return centroids


def median_score(coords, centroids, score_method):
    """

    :param coords:
    :param centroids:
    :param score_method:
    :return:
    """
    from sklearn.metrics import pairwise_distances
    from sklearn.preprocessing import normalize

    # n_rand = 10
    # for i in range(n_rand):
    #     randcoords = np.random.normal(size=coords.shape)
    #     randcoords = normalize(randcoords)
    #     dists = pairwise_distances(randcoords, centroids, metric='cosine')
    #     d2min = np.partition(dists, kth=2)[:, :2]
    #     if scoreMethod == 'median':
    #         b = np.median(d2min[:, 1])
    #         a = np.median(d2min[:, 0])
    #     else:
    #         b = np.mean(d2min[:, 1])
    #         a = np.mean(d2min[:, 0])
    #     r_score = b / a

    dists = pairwise_distances(coords, centroids, metric='cosine')
    cdist = pairwise_distances(centroids, centroids, metric='cosine')
    normal = cdist.mean()
    d2min = np.partition(dists, kth=1)[:, :2]

    a = d2min[:, 1]
    b = d2min[:, 0]
    s = a / b

    if score_method == 'median':
        score = np.median(s)
    else:
        score = np.mean(s)

    return score, s


def cluster_types(labels):
    """

    :param labels:
    :return:
    """
    _, counts = np.unique(labels, return_counts=True)
    thresh = 0.05 * np.mean(counts)
    counts = np.rint(counts / thresh) * thresh
    var = np.std(counts)
    ntypes = np.unique(counts).shape[0]
    return var, ntypes


def plotScores(pdb, n_range, scores, ntypes, save=True, save_path='score.png'):
    """

    :param pdb:
    :param n_range:
    :param scores:
    :param ntypes:
    """
    import matplotlib
    import matplotlib.pyplot as plt
    import numpy as np

    font = {'family': 'sans-serif',
            'weight': 'normal',
            'size': 10}

    # _, _, title = getPDB(pdb)
    title = pdb

    matplotlib.rc('font', **font)

    print('Plotting')
    fig, ax = plt.subplots(2, 1, figsize=(6, 3), sharex=True)
    fig.suptitle('Cluster quality score profile: ' + ' (' + pdb + ')')
    ax[0].scatter(n_range, scores, marker='D', s=15)
    ax[0].plot(n_range, scores)
    ax[1].plot(n_range, ntypes)
    ax[1].scatter(n_range, ntypes, marker='D', s=15)
    ax[0].axvline(x=n_range[np.argmax(scores)], label='Best Score = ' + str(n_range[np.argmax(scores)]), color='black')
    ax[1].axvline(x=n_range[np.argmax(scores)], label='Best Score', color='black')
    nc = str(n_range[np.argmax(scores)])
    ticks = ax[0].get_xticks()
    ticks = np.append(ticks, n_range[np.argmax(scores)])
    ax[1].set_xticks(ticks)
    ax[1].set_xlim([0, n_range[-1]])
    ax[1].set_xlabel('# Of Clusters')
    ax[0].set_ylabel('Quality' + '\n' + 'Score', rotation='horizontal', ha='center', va='center', labelpad=25)
    ax[1].set_ylabel('# Unique \n Clusters', rotation='horizontal', ha='center', va='center', labelpad=25)

    ax[0].tick_params(axis='y', labelsize=8)
    ax[1].tick_params(axis='y', labelsize=8)

    ax[0].grid()
    ax[1].grid()
    ax[0].legend()
    fig.tight_layout()
    if save:
        fig.savefig(f'{save_path}results_plot.svg')

    plt.show()


# Colormap generation taken from stackexchange
# https://stackoverflow.com/questions/42697933/colormap-with-maximum-distinguishable-colours

def generate_colormap(number_of_distinct_colors: int = 80):
    import math

    import numpy as np
    from matplotlib.colors import ListedColormap
    from matplotlib.cm import hsv

    if number_of_distinct_colors == 0:
        number_of_distinct_colors = 80

    number_of_shades = 7
    if number_of_distinct_colors < (number_of_shades - 1):
        number_of_shades = number_of_distinct_colors - 2

    number_of_distinct_colors_with_multiply_of_shades = int(
        math.ceil(number_of_distinct_colors / number_of_shades) * number_of_shades)

    # Create an array with uniformly drawn floats taken from <0, 1) partition
    linearly_distributed_nums = np.arange(
        number_of_distinct_colors_with_multiply_of_shades) / number_of_distinct_colors_with_multiply_of_shades

    # We are going to reorganise monotonically growing numbers in such way that there will be single array with saw-like pattern
    #     but each saw tooth is slightly higher than the one before
    # First divide linearly_distributed_nums into number_of_shades sub-arrays containing linearly distributed numbers
    arr_by_shade_rows = linearly_distributed_nums.reshape(number_of_shades,
                                                          number_of_distinct_colors_with_multiply_of_shades // number_of_shades)

    # Transpose the above matrix (columns become rows) - as a result each row contains saw tooth with values slightly higher than row above
    arr_by_shade_columns = arr_by_shade_rows.T

    # Keep number of saw teeth for later
    number_of_partitions = arr_by_shade_columns.shape[0]

    # Flatten the above matrix - join each row into single array
    nums_distributed_like_rising_saw = arr_by_shade_columns.reshape(-1)

    # HSV colour map is cyclic (https://matplotlib.org/tutorials/colors/colormaps.html#cyclic), we'll use this property
    initial_cm = hsv(nums_distributed_like_rising_saw)

    lower_partitions_half = number_of_partitions // 2
    upper_partitions_half = number_of_partitions - lower_partitions_half

    # Modify lower half in such way that colours towards beginning of partition are darker
    # First colours are affected more, colours closer to the middle are affected less
    lower_half = lower_partitions_half * number_of_shades
    for i in range(3):
        initial_cm[0:lower_half, i] *= np.arange(0.2, 1, 0.8 / lower_half)

    # Modify second half in such way that colours towards end of partition are less intense and brighter
    # Colours closer to the middle are affected less, colours closer to the end are affected more
    for i in range(3):
        for j in range(upper_partitions_half):
            modifier = np.ones(number_of_shades) - initial_cm[lower_half + j * number_of_shades: lower_half + (
                        j + 1) * number_of_shades, i]
            modifier = j * modifier / upper_partitions_half
            initial_cm[lower_half + j * number_of_shades: lower_half + (j + 1) * number_of_shades, i] += modifier

    return ListedColormap(initial_cm)

def getRGBA(values, rwb_scale):
    import matplotlib as mpl
    import numpy as np
    norm = mpl.colors.Normalize(vmin=np.min(values), vmax=np.max(values))
    if rwb_scale == 'True':
        import matplotlib.pyplot as plt
        cmap = plt.get_cmap('coolwarm_r')
        rgba = cmap(norm(values)) * 255
    else:
        cmap = generate_colormap(int(np.max(values)))
        rgba = cmap(norm(values)) * 255
    return rgba

from chimerax.core.commands import *
from chimerax.core import objects
from chimerax.std_commands import *
from chimerax.core.commands import all_objects
from chimerax.core.commands import run
from chimerax.std_commands import show
from chimerax.std_commands import hide
from chimerax.std_commands import select
from chimerax.atomic.molsurf import MolecularSurface
from chimerax.atomic import *
import numpy as np
import matplotlib as mpl
# Colormap taken from stackexchange

# from chimerax.core.commands import measure


from numpy.linalg import norm
import sys
print(sys.argv)

import argparse
import os

parser = argparse.ArgumentParser(description='ChimeraX script for visualization of pyCapsid results')
parser.add_argument('-report_dir', help='Location of pyCapsid report chimerax directory. Only needed if this script is in a separate directory.', default=None, required=False)
parser.add_argument('-nc', help='Number of clusters in the corresponding results you want to visualize', default=None, required=False)
parser.add_argument('-remote', help='Whether to use a remote structure from the PDB database', default=None, required=False)
parser.add_argument('-mode', help='Whether to end on one visualization or the other, or create a duplicate model for the cluster results and quality score results for further visualization. Options are: cluster, score, both', default='cluster', required=False)
parser.add_argument('-addClusterIDs', help='Whether to add custom cluster_id attr to each residue', default=False, required=False)
parser.add_argument('-allClusters', help='Whether to create an individual snapshot of each cluster', default=False, required=False)
#parser.add_argument('-clusterChains', help='Whether to change chain_ids to reflect clusters', default=False, required=False)
parser.add_argument('-pdb', help='If remote is True or none, PDBID of the target structure. Otherwise, the local filename of the target structure', default=None, required=False)
args = vars(parser.parse_args())

if args['report_dir'] is None:
    report_dir = os.path.dirname(sys.argv[0])
    print(report_dir)
else:
    report_dir = args['report_dir']
print(report_dir)
os.chdir(report_dir)


if args['pdb'] is None:
    fs = [x for x in os.listdir('./') if x.endswith('_final_results.npz')]
    f = fs[0]
    pdb = f.split('_final')[0]
else:
    pdb = args['pdb']

if args['remote'] is None:
    files = [x for x in os.listdir('./') if (x.endswith('.pdb') or x.endswith('.cif'))]
    if len(files) == 0:
        remote = 'True'
    elif len(files) == 1:
        remote = 'False'
        pdb = files[0]
    else:
        raise Exception("More than one .pdb file in directory")
else:
    remote = args['remote']

if args['nc'] is None:
    n_clusters = np.load(f'{pdb}_final_results.npz')['nc']
    nc_filename = ''
else:
    n_clusters = int(args['nc'])
    nc_filename = f'_{n_clusters}'

vis_mode = args['mode']
all_clusters = args['allClusters']
#cluster_models = args['clusterChains']
add_cluster_id = args['addClusterIDs']

run(session, 'set bg white')
run(session, 'graphics silhouettes true')

if remote=='True':
    run(session, f'open {pdb}')
    run(session, 'view orient')
    run(session, 'lighting soft')
    run(session, 'hide atoms')
    run(session, 'show cartoons')
    run(session, f'save ../figures/structures/{pdb}_asymmetric_unit.png')
    run(session, 'sym #1 assembly 1 copies True')
    run(session, 'close #1')
    run(session, 'view orient')
    run(session, 'hide atoms')
    run(session, 'show cartoons')
    run(session, f'save ../figures/structures/{pdb}_full_capsid.png')
    run(session, 'rename #2 id #1')
else:
    run(session, f'open {pdb}')
    run(session, 'view orient')
    run(session, 'lighting soft')
    run(session, 'hide atoms')
    run(session, 'show cartoons')
    run(session, f'save ../figures/structures/{pdb}_asymmetric_unit.png')
    run(session, f'save ../figures/structures/{pdb}_full_capsid.png')

# center of mass measurement from ChimeraX tutorial
run(session, 'sel protein')
run(session, 'del ~sel')
run(session, 'sel clear')
if vis_mode == 'both':
    run(session, f'combine #1 close true modelId 1 name {pdb}_copy')
atoms = all_objects(session).atoms  # getting atom list
coords = atoms.scene_coords  # getting atom coords
modelCenter = coords.mean(axis=0)  # calculate center of mass
print("modelCenter", modelCenter)

if modelCenter.any:
    print('Aligning Models')
    modelX, modelY, modelZ = modelCenter
    run(session, 'move x ' + str(-1 * modelX) + ' coordinateSystem #1 models #1')  # adjust x coordinates
    run(session, 'move y ' + str(-1 * modelY) + ' coordinateSystem #1 models #1')  # adjust y coordinates
    run(session, 'move z ' + str(-1 * modelZ) + ' coordinateSystem #1 models #1')  # adjust z coordinates
    run(session, 'view orient')
else:
    print('Models Already Aligned')

coords = atoms.scene_coords
radius = norm(coords, axis=1).max()  # calculate the norms of the x coordinates, and then choose the maximum value
print("radius_est: ", radius)


results = np.load(f'{pdb}_final_results_full.npz')
labels = results['labels']
nc_range = results['nc_range']
scores = results['full_scores']
print(nc_range)
print(n_clusters)
ind = np.argwhere(nc_range == n_clusters)[0][0]
print(ind)

print(f'Visualizing cluster results of {pdb} for {n_clusters} clusters')
labels = labels[ind]
score = scores[ind]



rgba_scores = getRGBA(score, rwb_scale='True')
rgba_clusters = getRGBA(labels, rwb_scale='False')

# Score visualization
residues = atoms.unique_residues
cx_nr = len(residues)
enm_nr = rgba_scores.shape[0]

print('# of residues:', cx_nr)
print('# of ENM residues:', enm_nr)

if add_cluster_id:
    label_nums = np.unique(labels)
    clusters_residues = []
    for num in label_nums:
        c_index = np.where(labels == num)[0]
        cluster_residues = []
        # Gather cluster residues in a list
        for i in c_index:
            cluster_residues.append(residues[i])
        clusters_residues.append(cluster_residues)
    for i, clust in enumerate(clusters_residues):
        cluster_obj = objects.Objects()
        for residue_obj in clust:
            cluster_obj.add_atoms(residue_obj.atoms)
        select.select(session, cluster_obj, residues=True)
        run(session, f'setattr sel residues cluster_id {i} create true')


if vis_mode == 'score':
    rgba_one = rgba_clusters
    rgba_two = rgba_scores
    filename_1 = 'highest_quality_clusters'
    filename_2 = 'residue_cluster_scores'
elif vis_mode == 'clusters':
    rgba_one = rgba_scores
    rgba_two = rgba_clusters
    filename_2 = 'highest_quality_clusters'
    filename_1 = 'residue_cluster_scores'
else:
    rgba_one = rgba_scores
    rgba_two = rgba_clusters
    filename_2 = 'highest_quality_clusters'
    filename_1 = 'residue_cluster_scores'


if all_clusters:
    # Color clusters normally
    for i in range(cx_nr):
        res = residues[i]
        res.ribbon_color = rgba_one[i, :]
        ats = res.atoms
        for at in ats:
            at.color = rgba_one[i, :]
    # hide everything
    run(session, 'hide all ribbons')
    # Show each cluster individually
    # Get indices of each cluster
    label_nums = np.unique(labels)
    clusters_residues = []
    for num in label_nums:
        c_index = np.where(labels == num)[0]
        cluster_residues = []
        # Gather cluster residues in a list
        for i in c_index:
            cluster_residues.append(residues[i])
        clusters_residues.append(cluster_residues)
    #print(cluster_residues)
    for i, clust in enumerate(clusters_residues):
        cluster_obj = objects.Objects()
        for residue_obj in clust:
            cluster_obj.add_atoms(residue_obj.atoms)
        show.show(session, objects=cluster_obj, target={'models', 'cartoons'})
        run(session, 'view orient')
        run(session, f'save ../figures/structures/{pdb}_{filename_1}{nc_filename}_cluster_{i}.png')
        hide.hide(session, cluster_obj, target={'models', 'cartoons'})
    run(session, 'show all cartoons')
    run(session, 'select clear')
else:
    for i in range(cx_nr):
        res = residues[i]
        res.ribbon_color = rgba_one[i, :]
        ats = res.atoms
        for at in ats:
            at.color = rgba_one[i, :]

run(session, 'view orient')
run(session, f'save ../figures/structures/{pdb}_{filename_1}{nc_filename}.png')

# Cluster visualization
if vis_mode == 'both':
    run(session, 'combine #1 name scores')
    run(session, 'hide #2 models')


if all_clusters:
    for i in range(cx_nr):
        res = residues[i]
        res.ribbon_color = rgba_two[i, :]
        ats = res.atoms
        for at in ats:
            at.color = rgba_two[i, :]
    # hide everything
    run(session, 'hide all ribbons')
    # Show each cluster individually
    # Get indices of each cluster
    label_nums = np.unique(labels)
    clusters_residues = []
    for num in label_nums:
        c_index = np.where(labels == num)[0]
        cluster_residues = []
        # Gather cluster residues in a list
        for i in c_index:
            cluster_residues.append(residues[i])
        clusters_residues.append(cluster_residues)
    # print(cluster_residues)
    for i, clust in enumerate(clusters_residues):
        cluster_obj = objects.Objects()
        for residue_obj in clust:
            cluster_obj.add_atoms(residue_obj.atoms)
        show.show(session, objects=cluster_obj, target={'models', 'cartoons'})
        run(session, 'view orient')
        run(session, f'save ../figures/structures/{pdb}_{filename_2}{nc_filename}_cluster_{i}.png')
        hide.hide(session, cluster_obj, target={'models', 'cartoons'})
    run(session, 'show all cartoons')
else:
    for i in range(cx_nr):
        res = residues[i]
        res.ribbon_color = rgba_two[i, :]
        ats = res.atoms
        for at in ats:
            at.color = rgba_two[i, :]

run(session, 'view orient')
run(session, f'save ../figures/structures/{pdb}_{filename_2}{nc_filename}.png')




import numpy as np
import numba as nb


# @nb.njit()
def fastFlucts(evals, evecs, n_modes, is3d):
    """

    :param evals:
    :param evecs:
    :param n_modes:
    :param is3d:
    :return:
    """
    n_d = evecs.shape[0]
    flucts = np.zeros(n_d)
    for i in range(n_modes):
        flucts += 1 / evals[i] * evecs[:, i] ** 2
    if is3d:
        return np.reshape(flucts, (-1, 3)).sum(axis=-1)
    else:
        return flucts


# @nb.njit()
def checkIcoFlucts(flucts):
    """

    :param flucts:
    :return:
    """
    F = np.reshape(flucts, (60, -1))
    devs = np.ptp(F, axis=0)
    d = np.max(devs)
    # print('Maximum deviation from icosahedral: ', np.max(devs))
    return d


def springFit(bfactors, sqFlucts):
    """

    :param bfactors:
    :param sqFlucts:
    :return:
    """
    import statsmodels.api as sm

    intercept = False
    if intercept:
        sqFlucts = sm.add_constant(sqFlucts)
    M = sm.RLM(bfactors, sqFlucts, M=sm.robust.norms.HuberT())
    results = M.fit()
    # print(results.summary(xname=['Squared Fluctuations'], yname='B-factors', title='Results of fitting predicted squared fluctuations'))
    print(results.summary2(xname=['Squared Fluctuations'], yname='B-factors',
                          title='Summary of regression results: predicted squared fluctuations vs B-factors'))
    a = results.params[-1]
    stderr = results.bse * np.sqrt(bfactors.shape[0])
    pv = results.pvalues
    ci = results.conf_int(alpha=0.05)

    if intercept:
        b = results.params[0]
    else:
        b = 0

    # from sklearn.linear_model import HuberRegressor
    #
    # huber = HuberRegressor(fit_intercept=False, tol=0, alpha=0.0).fit(sqFlucts, bfactors)
    # a = huber.coef_
    # b = huber.intercept_

    return a, b, stderr, ci, pv


def fluctModes(evals, evecs, bfactors, is3d, isIco):
    """

    :param evals:
    :param evecs:
    :param bfactors:
    :param is3d:
    :return:
    """
    coeffs = []
    ico_devs = []

    for n_modes in range(1, len(evals)):
        flucts = fastFlucts(evals, evecs, n_modes, is3d)
        cc = np.corrcoef(bfactors, flucts)[1, 0]
        if isIco:
            icodev = checkIcoFlucts(flucts)
        else:
            icodev = 0
        coeffs.append(cc)
        ico_devs.append(icodev)
    return coeffs, ico_devs


def fitBfactors(evals, evecs, bfactors, is3d, isIco=True, fitModes=False, plotModes=False, forceIco=False, icotol=0.002, save_path='./'):
    """

    :param evals:
    :param evecs:
    :param bfactors:
    :param is3d:
    :param fitModes:
    :param plotModes:
    :param forceIco:
    :param icotol:
    :return:
    """
    n_modes = evals.shape[0]
    if np.var(bfactors)==0:
        print('Warning! Experimental b-factors have zero variance. pyCapsid will add random noise to b-factors to calculate a spring constant. The CC values will not be accurate and all modes will be used.')
        fitModes = False
        forceIco = False

    if plotModes:
        coeffs, ico_devs = fluctModes(evals, evecs, bfactors, is3d, isIco)
        plotByMode(np.arange(1, n_modes), coeffs, 'CC', save_path=save_path)
        if fitModes:
            if forceIco:
                plotByMode(np.arange(1, n_modes), ico_devs, 'Icosahedral Deviation', save_path=save_path)
                icoI = np.nonzero(np.array(ico_devs) < icotol)
                n_m = np.argmax(np.array(coeffs)[icoI])
                coeff = np.array(coeffs)[icoI][n_m]
                ico_dev = np.array(ico_devs)[icoI][n_m]
            else:
                n_m = np.argmax(coeffs)
                coeff = coeffs[n_m]
                ico_dev = ico_devs[n_m]
        else:
            n_m = n_modes - 1
            flucts = fastFlucts(evals, evecs, n_m, is3d)
            coeff = np.corrcoef(bfactors, flucts)[1, 0]
            if isIco:
                ico_dev = checkIcoFlucts(flucts)
            else:
                ico_dev = 0
        flucts = fastFlucts(evals, evecs, n_m, is3d)
    else:
        n_m = n_modes - 1
        flucts = fastFlucts(evals, evecs, n_m, is3d)
        coeff = np.corrcoef(bfactors, flucts)[1, 0]
        if isIco:
            ico_dev = checkIcoFlucts(flucts)
        else:
            ico_dev = 0

    k, intercept, stderr, ci, pv = springFit(bfactors, flucts[:, np.newaxis])

    bfactors_predicted = k * flucts + intercept

    return coeff, k, intercept, bfactors_predicted, ci, pv, ico_dev, n_m


def plotByMode(mode_indices, data, datalabel, save_path='./'):
    """

    :param mode_indices:
    :param data:
    :param datalabel:
    """
    import matplotlib.pyplot as plt
    import numpy as np
    fig, ax = plt.subplots(1, 1)
    ax.plot(mode_indices, data)
    # ax[0].vlines(nModes, np.min(coeffs), np.max(coeffs))
    ax.set_xlabel('Number Of Modes')
    ax.set_ylabel(datalabel)
    fig.suptitle(datalabel + ' vs number of low frequency modes')
    plt.show()
    fig.savefig(f'{save_path}{datalabel}_by_mode.svg')
    np.savez(f'{save_path}{datalabel}_by_mode.npz', mode_indices = mode_indices, data = data)


def fitPlotBfactors(evals, evecs, bfactors, pdb, is3d=True, fitModes=True, plotModes=False, forceIco=True, icotol=0.002,
                    save=True, save_path='./', isIco=False):
    """

    :param isIco:
    :param evals:
    :param evecs:
    :param bfactors:
    :param pdb:
    :param is3d:
    :param fitModes:
    :param plotModes:
    :param forceIco:
    :param icotol:
    :return:
    """
    coeff, k, intercept, bfactors_predicted, ci, pv, ico_dev, nmodes = fitBfactors(evals, evecs, bfactors, is3d, isIco,
                                                                                   fitModes, plotModes, forceIco,
                                                                                   icotol, save_path=save_path)

    print(f'Number of low-frequency modes used: {nmodes}')
    k_ci = np.abs(ci[0][0] - ci[0][1])
    print(f'Scale factor k between predicted fluctuations and B-factors: {k:.2e}±{k_ci:.2e}')
    gamma = (8 * np.pi ** 2) / k
    ci = (8 * np.pi ** 2) / ci

    if is3d:
        gamma = gamma / 3
        ci = ci / 3

    gamma_ci = np.abs(ci[0][0] - ci[0][1])
    print(f'Estimated spring constant gamma of ENM springs: {gamma:.2e}±{gamma_ci:.2e}')

    file = f'{save_path}b_factors.npz'
    print('Saving B-factor results in' + file)
    residue_numbers = np.arange(1, bfactors.shape[0] + 1)
    np.savez_compressed(file, bfactors_predicted=bfactors_predicted, bfactors_experimental=bfactors, CC=coeff, gamma=gamma, gamma_ci=gamma_ci, k=k, k_ci=k_ci, n_modes=nmodes, n_asym=int(bfactors.shape[0] / 60), residue_numbers=residue_numbers)
    
    import matplotlib.pyplot as plt
    import matplotlib
    fig, ax = plt.subplots(1, 1, figsize=(6, 3))
    font = {'family': 'sans-serif',
            'weight': 'normal',
            'size': 9}
    matplotlib.rc('font', **font)

    if isIco:
        n_asym = int(bfactors.shape[0] / 60)
        plotx_experimental = np.arange(bfactors.shape[0])[:n_asym]
        ploty_experimental = bfactors[:n_asym]
        plotx_predicted = np.arange(bfactors_predicted.shape[0])[:n_asym]
        ploty_predicted = bfactors_predicted[:n_asym]
    else:
        plotx_experimental = np.arange(bfactors.shape[0])
        ploty_experimental = bfactors
        plotx_predicted = np.arange(bfactors_predicted.shape[0])
        ploty_predicted = bfactors_predicted

    ax.plot(plotx_experimental, ploty_experimental, label='b-factors (Experimental)')
    ax.plot(plotx_predicted, ploty_predicted, label='b-factors (Predicted)')
    ax.set_ylabel(r'$Å^{2}$', fontsize=9)
    ax.set_xlabel('Residue Number', fontsize=9)
    ax.tick_params(axis='y', labelsize=8)
    ax.tick_params(axis='x', labelsize=8)

    ax.legend()
    fig.suptitle(
        (fr"Experimental vs Predicted b-factors: ({pdb})" + '\n' + fr"$\gamma = $ {gamma:.2e} $\pm$ {gamma_ci:.2e} "
                                                                   fr"$\frac{{k_b T}}{{Å^{2}}}$ CC = {coeff:.2f} Icosahedral deviation = {ico_dev:.2f}"),
        fontsize=9)
    fig.tight_layout()
    if save:
        fig.savefig(f'{save_path}b_factors.svg')
    plt.show()

    return coeff, gamma, nmodes



import numpy as np
def generateTrajFromModes(atoms, modes, evals, n_mode=0, max_amplitude=None, n_frames = 20):
    residues = atoms.unique_residues
    coords = atoms.scene_coords

    mode_vec = modes[:, n_mode].reshape((-1, 3))
    if max_amplitude is not None:
        vector_lengths = np.sqrt(np.sum(mode_vec**2, axis=-1))
        scale = max_amplitude / np.max(vector_lengths)
        mode_vec *= scale
    else:
        mode_vec *= 1/evals[n_mode]


    time = np.linspace(0, 2*np.pi, n_frames, endpoint=False)
    deviation = np.sin(time)[:, np.newaxis, np.newaxis] * mode_vec
    print(deviation.shape)
    oscillation = np.zeros((n_frames, coords.shape[0], 3))

    ind = 0
    for i in range(len(residues)):
        res = residues[i]
        n_atoms = res.num_atoms
        res_start = ind
        res_stop = ind + n_atoms - 1
        ind += n_atoms
        dev = deviation[:,i:i+1,:]
        oscillation[:,res_start:res_stop,:] = coords[res_start:res_stop,:] + dev
    return oscillation

from chimerax.core.commands import *
from chimerax.std_commands import *
from chimerax.core.commands import all_objects
from chimerax.core.commands import run
from chimerax.atomic.molsurf import MolecularSurface
from chimerax.atomic import *
import matplotlib as mpl
# Colormap taken from stackexchange

# from chimerax.core.commands import measure


from numpy.linalg import norm
import sys
print(sys.argv)

import argparse
import os

parser = argparse.ArgumentParser(description='ChimeraX script for visualization of pyCapsid results')
parser.add_argument('-report_dir', help='Location of pyCapsid report chimerax directory', default=None, required=False)
parser.add_argument('-nmode', help='Index of normal mode to visualize', default=None, required=False)
parser.add_argument('-remote', help='Whether to use a remote structure from the PDB database', default=None, required=False)
parser.add_argument('-pdb', help='If remote is True or none, PDBID of the target structure. Otherwise, the local filename of the target structure', default=None, required=False)
parser.add_argument('-amplitude', help='Maximum amplitude of motion along the mode for visualization purposes', default=None, required=False)
parser.add_argument('-frames', help='Number of frames to generate for the animation', default=20, required=False)
parser.add_argument('-save_frames', help='Whether to save the individual frames used to create the image', default=False, required=False)
args = vars(parser.parse_args())

if args['report_dir'] is None:
    report_dir = os.path.dirname(sys.argv[0])
    print(report_dir)
else:
    report_dir = args['report_dir']
print(report_dir)
os.chdir(report_dir)

if args['pdb'] is None:
    fs = [x for x in os.listdir('./') if x.endswith('_final_results.npz')]
    f = fs[0]
    pdb = f.split('_final')[0]
else:
    pdb = args['pdb']

if args['remote'] is None:
    files = [x for x in os.listdir('./') if (x.endswith('.pdb') or x.endswith('.cif'))]
    if len(files) == 0:
        remote = 'True'
    elif len(files) == 1:
        remote = 'False'
        pdb = files[0]
    else:
        raise Exception("More than one .pdb file in directory")
else:
    remote = args['remote']

if args['amplitude'] is None:
    max_amplitude = 3
else:
    max_amplitude = float(args['amplitude'])

if args['nmode'] is None:
    modes = np.load(f'{pdb}_modes.npz')
    evals = modes['eigen_vals']
    uniques, inds, counts = np.unique(evals.round(decimals=8), return_index=True, return_counts=True)
    icoEvalInds = inds[counts == 1]
    print('Indices of non-degenerate eigenmodes: ', icoEvalInds)
    if len(icoEvalInds) == 0:
        print('No non-degenerate modes found')
        n_mode = 0
    else:
        print('Visualizing lowest frequency non-degenerate mode')
        n_mode=icoEvalInds[0]
else:
    n_mode = int(args['nmode'])

if args['frames'] is None:
    n_frames = 20
else:
    n_frames = int(args['frames'])

save_frames = args['save_frames']

run(session, 'set bg white')
run(session, 'graphics silhouettes true')

if remote=='True':
    run(session, f'open {pdb}')
    run(session, 'sym #1 assembly 1 copies True')
    run(session, 'close #1')
    run(session, 'view orient')
    run(session, 'hide atoms')
    run(session, 'show cartoons')
    run(session, 'lighting soft')
    run(session, f'save ../figures/structures/{pdb}_full_capsid.png')
    run(session, 'rename #2 id #1')
else:
    run(session, f'open {pdb}')
    run(session, 'view orient')
    run(session, 'lighting soft')
    run(session, 'hide atoms')
    run(session, 'show cartoons')
    run(session, f'save ../figures/structures/{pdb}_full_capsid.png')

# center of mass measurement from ChimeraX tutorial
run(session, 'sel protein')
run(session, 'del ~sel')
run(session, 'sel clear')
atoms = all_objects(session).atoms  # getting atom list
coords = atoms.scene_coords  # getting atom coords
modelCenter = coords.mean(axis=0)  # calculate center of mass
print("modelCenter", modelCenter)

if modelCenter.any:
    print('Aligning Models')
    modelX, modelY, modelZ = modelCenter
    run(session, 'move x ' + str(-1 * modelX) + ' coordinateSystem #1 models #1')  # adjust x coordinates
    run(session, 'move y ' + str(-1 * modelY) + ' coordinateSystem #1 models #1')  # adjust y coordinates
    run(session, 'move z ' + str(-1 * modelZ) + ' coordinateSystem #1 models #1')  # adjust z coordinates
    run(session, 'view orient')
else:
    print('Models Already Aligned')

coords = atoms.scene_coords

modes = np.load(f'{pdb}_modes.npz')
evecs = modes['eigen_vecs']
evals = modes['eigen_vals']

atoms = all_objects(session).atoms
osc = generateTrajFromModes(atoms, evecs, evals, n_mode=n_mode, max_amplitude=max_amplitude, n_frames=n_frames)

print(f'Visualizing motion of {pdb} structure along mode # {n_mode}')
run(session, 'view orient pad -0.2')
run(session, 'color bychain')
run(session, 'movie record supersample 3 directory ./ format png')

from time import sleep
for i in range(n_frames):
    atoms.scene_coords = osc[i, :, :]
    run(session, 'wait 1')
    #run(session, f'save ../figures/structures/{pdb}_mode_{n_mode}_animation_frame_{i}.png')
if save_frames:
    run(session, f'movie encode quality high resetMode keep output ../figures/structures/{pdb}_mode_{n_mode}_animation.mp4')
else:
    run(session, f'movie encode quality high output ../figures/structures/{pdb}_mode_{n_mode}_animation.mp4')





"""Module with functions for building hessian matrices of different elastic network models."""

import numba as nb
import numpy as np

def buildENMPreset(coords, preset='ANM', **kwargs):
    """Builds a hessian matrix representing an ENM based on one of several presets.

    :param ndarray coords: Cartesian coordinates of alpha carbon atoms(or choice of alternate representation).
    :param str preset: The specific model preset to use. Only accepts the following values:
        - 'ANM': Anisotropic Network Model with a cutoff of 15Å and no distance weighting.
        - 'GNM': Gaussian Network Model with a cutoff of 7.5Å and no distance weighting.
        - 'U-ENM': Unified Elastic Network Model with a cutoff of 7.5Å and and f_anm parameter of 0.1. [ref]
        - 'bbENM': Backbone-enhanced Elastic Network Model with a cutoff of 7.5Å and no distance weighting.
        - 'betaENM': Not yet implemented
    :return: A tuple of sparse matrices. The kirchoff matrix and the hessian matrix
    :rtype: (scipy.sparse.csr_matrix, scipy.sparse.csr_matrix)
    """
    model_presets = ['ANM', 'GNM', 'U-ENM', 'bbENM', 'custom']
    print('Building hessian for model preset: ', preset)
    if preset == 'ANM':
        cutoff = 15
        return buildENM(coords, cutoff=cutoff, **kwargs)
    elif preset == 'GNM':
        cutoff = 7.5
        gnm=True
        return buildENM(coords, cutoff=cutoff, gnm=gnm, **kwargs)
    elif preset == 'U-ENM':
        cutoff = 7.5
        fanm = 0.1
        return buildENM(coords, cutoff=cutoff, fanm=fanm, **kwargs)
    elif preset == 'bbENM':
        cutoff = 7.5
        l_backbone=1
        k_backbone = 10
        if 'chain_starts' not in kwargs:
            raise ValueError("No chain information provided. Indices of chain starts must be provided as chain_starts")
        chain_starts = kwargs.pop('chain_starts')
        return buildENM(coords, cutoff=cutoff, chain_starts=chain_starts, l_backbone=l_backbone, k_backbone=k_backbone, **kwargs)
    elif preset == 'custom':
        print('No preset specified, using defaults and whatever specific parameters were specified.')
        return buildENM(coords, **kwargs)
    else:
        raise ValueError("Invalid model preset. Expected one of: %s" % model_presets)



def buildENM(coords, cutoff=10, gnm=False, fanm=1, wfunc='power', base_dist=1, d_power=0, backbone=False, k_backbone=1,
             l_backbone=1, chain_starts=None, save_hessian=False, save_kirchhoff=False, save_cg_path='./'):
    """Builds a hessian matrix representing an ENM based on the provided parameters.

        :param coords: Cartesian of alpha carbon (or choice of representation) atoms
        :param float cutoff: Cutoff distance for long range interactions in the ENM
        :param gnm: Whether to use only the kirchhoff matrix (Isotropic GNM), otherwise use full hessian
        :param float fanm: Parameter representing degree of anisotropy for U-ENM
        :param wfunc: Weight function to assign spring constant based on distance. Use 'power' or 'exp'
        :param base_dist: In wfunc, divide distance by the base_distance
        :param d_power: In wfunc, use this power of the distance
        :param backbone: Whether to use stronger interactions for residues connected along the backbone
        :param k_backbone: Relative strength of backbone interaction
        :param l_backbone: How many steps along the backbone to give stronger interactions
        :param chain_starts: Used for defining backbone interactions
        :return: A tuple of sparse matrices. The kirchhoff matrix and the hessian matrix
        :rtype: (scipy.sparse.csr_matrix, scipy.sparse.csr_matrix)
        """
    params = locals()
    params.pop('coords')
    print('Model parameters: ', params)

    import numpy as np
    from scipy import sparse
    from sklearn.neighbors import BallTree, radius_neighbors_graph


    n_atoms = coords.shape[0]
    dof = n_atoms * 3

    print(f'Finding neighbors within {cutoff}Å')
    distGraph = radius_neighbors_graph(coords, cutoff, mode='distance', n_jobs=-1)
    dists = distGraph.tocoo().copy()
    dists.sum_duplicates()

    print('Building kirchhoff matrix')
    kirch = kirchGamma(dists, gfunc=wfunc, bd=base_dist, d2=d_power).tolil()

    if backbone:
        print('Adding backbone terms')
        try:
            c = chain_starts[0]
        except:
            print('Must provide an array of chain starts')
        buildBackbone(l_backbone, k_backbone, kirch, chain_starts)

    dg = np.array(kirch.sum(axis=0))
    kirch.setdiag(-dg[0])
    kirch = kirch.tocsr()

    if not gnm:
        print('Building hessian matrix')
        kc = kirch.tocoo().copy()
        hData = hessCalc(kc.row, kc.col, kc.data, coords)
        indpt = kirch.indptr
        inds = kirch.indices
        hessian = sparse.bsr_matrix((hData, inds, indpt), shape=(dof, dof)).tocsr()
        hessian = fanm * hessian + (1 - fanm) * sparse.kron(kirch, np.identity(3))
    else:
        hessian = kirch.copy()

    print('Done building model')

    from sklearn.utils.validation import check_symmetric
    check_symmetric(hessian, raise_warning=True, tol=1e-5)
    check_symmetric(kirch, raise_warning=True, tol=1e-5)
    hessian.eliminate_zeros()
    kirch.eliminate_zeros()

    if save_kirchhoff:
        file = save_cg_path + "kirchhoff"
        sparse.save_npz(file, kirch, compressed=True)

    if save_hessian:
        file = save_cg_path + "hessian"
        sparse.save_npz(file, hessian, compressed=True)

    return kirch, hessian


def kirchGamma(dists, **kwargs):
    """ Calculates the kirchhoff matrix (spring constant matrix) from a sparse distance matrix.

    :param dists: Sparse distance matrix
    :param kwargs:
    :return: The sparse kirchhoff matrix.
    """
    if kwargs['gfunc'] == 'power':
        dists.data = -1 / ((dists.data / kwargs['bd']) ** kwargs['d2'])
    elif kwargs['gfunc'] == 'exp':
        dists.data = -np.exp(((dists.data / kwargs['bd']) ** kwargs['d2']))
    else:
        dists.data = -np.ones_like(dists.data)

    return dists


def buildBackbone(bblen, bbgamma, kirch, chain_starts):
    """ Modifies the kirchhoff matrix to include backbone terms

    :param bblen: The number of backbone neighbors in either direction with strengthened spring constants.
    :param bbgamma: The relative strength to use for the backbone interactions.
    :param kirch: The sparse kirchhoff matrix
    :param chain_starts: Indices where protein chains begin/end.
    """
    for i in range(len(chain_starts) - 1):
        start = chain_starts[i]
        stop = chain_starts[i + 1]
        for j in range(stop - start):
            ij = start + j
            for k in range(1, bblen + 1):
                neighbor = ij - k
                if neighbor < start:
                    continue
                kirch[ij, neighbor] = -bbgamma / k
                kirch[neighbor, ij] = -bbgamma / k


@nb.njit()
def hessCalc(row, col, kGamma, coords):
    """
    :param row:
    :param col:
    :param kGamma:
    :param coords:
    :return:
    """
    hData = np.zeros((row.shape[0], 3, 3))
    dinds = np.nonzero(row == col)[0]
    for k, (i, j) in enumerate(zip(row, col)):
        if i == j:
            continue
        dvec = coords[j] - coords[i]
        d2 = np.dot(dvec, dvec)
        g = kGamma[k]
        hblock = np.outer(dvec, dvec) * (g / d2)
        hData[k, :, :] = hblock
        hData[dinds[i]] += -hblock
        # hData[dinds[j]] += -hblock
    return hData


# @nb.njit()
# def cooOperation(row, col, data, func, arg):
#     r = np.copy(data)
#     for n, (i, j, v) in enumerate(zip(row, col, data)):
#         if i == j:
#             continue
#         r[n] = func(i, j, v, arg)
#     return r


# def betaCarbonModel(calphas):
#     """
#
#     :param calphas:
#     :return:
#     """
#     from sklearn.neighbors import BallTree, radius_neighbors_graph
#     from scipy import sparse
#     coords = calphas.getCoords()
#     n_atoms = coords.shape[0]
#     n_asym = int(n_atoms / 60)
#
#     print('# Atoms ', n_atoms)
#     dof = n_atoms * 3
#     betaCoords, bvals, nobetas, aBetas = buildBetas(coords, calphas)
#     print(bvals.shape)
#     print(betaCoords.shape)
#
#     tree = BallTree(coords)
#     kirch = radius_neighbors_graph(tree, cutoff, mode='distance', n_jobs=-1)
#     kc = kirch.tocoo()
#     kc.sum_duplicates()
#     kc.eliminate_zeros()
#     bfactors = calphas.getBetas()
#     kirch = kirchGamma(kc, d2=d2)
#     if backboneConnect:
#         kbb = backboneStrength
#         kirch = backbonePrody(calphas, kirch.tolil(), kbb, s=bblen)
#     print('nonzero values: ', kirch.nnz)
#     dg = np.array(kirch.sum(axis=0))
#     kirch.setdiag(-dg[0])
#     kirch.sum_duplicates()
#     kirch = kirch.tocsr()
#
#     btree = BallTree(betaCoords)
#     betaKirch = radius_neighbors_graph(btree, cutoff, mode='distance', n_jobs=-1).tocoo()
#     betaKirch.eliminate_zeros()
#     betaKirch = kirchGamma(betaKirch.tocoo(), d2=d2)
#     dg = np.array(betaKirch.sum(axis=0))
#     betaKirch.setdiag(-dg[0])
#     betaKirch = betaKirch.tocsr()
#
#     from scipy.spatial import KDTree
#     akdtree = KDTree(coords)
#     bkdtree = KDTree(betaCoords)
#     abKirch = akdtree.sparse_distance_matrix(bkdtree, cutoff).tocoo()
#     abKirch = kirchGamma(abKirch.tocoo(), d2=d2)
#     dg = np.array(abKirch.sum(axis=0))
#     abKirch.setdiag(-dg[0])
#     abKirch.eliminate_zeros()
#     abKirch = abKirch.tocoo()
#
#     print('aa Hess')
#     kc = kirch.tocoo().copy()
#     kc.sum_duplicates()
#     haaData = hessCalc(kc.row, kc.col, kirch.data, coords)
#     indpt = kirch.indptr
#     inds = kirch.indices
#     haa = sparse.bsr_matrix((haaData, inds, indpt), shape=(dof, dof)).tolil()
#     haa = fanm * haa + (1 - fanm) * sparse.kron(kirch, np.identity(3))
#     haa = haa.tocsr()
#     haa.eliminate_zeros()
#     haa.sum_duplicates()
#
#     print('bb Hess')
#     kbc = betaKirch.tocoo().copy()
#     kbc.sum_duplicates()
#     row, col = (kbc.row, kbc.col)
#     hbrow, hbcol, hbdata, kbrow, kbcol, kbdata = bbHess(row, col, betaCoords, nobetas, bvals, aBetas, betaKirch.data,
#                                                         fanm)
#     hbb = sparse.coo_matrix((hbdata, (hbrow, hbcol)), shape=(dof, dof)).tocsr()
#     kbb = sparse.coo_matrix((kbdata, (kbrow, kbcol)), shape=(n_atoms, n_atoms)).tocsr()
#     hbb.eliminate_zeros()
#     hbb.sum_duplicates()
#     kbb.eliminate_zeros()
#     kbb.sum_duplicates()
#
#     print('ab Hess')
#     # abKirch = kirchGamma(abKirch.tocoo(), bfactors, d2=d2).tocoo()
#     abKirch.eliminate_zeros()
#     abKirch.sum_duplicates()
#     row, col, dat = (abKirch.row, abKirch.col, abKirch.data)
#
#     hrow, hcol, habData, krow, kcol, kabData = abHessOnly(row, col, dat, betaCoords, coords, nobetas, bvals, aBetas,
#                                                           fanm)
#     hab = sparse.coo_matrix((habData, (hrow, hcol)), shape=(dof, dof)).tocsr()
#     kab = sparse.coo_matrix((kabData, (krow, kcol)), shape=(n_atoms, n_atoms)).tocsr()
#
#     from sklearn.utils.validation import check_symmetric
#     check_symmetric(haa, raise_exception=True, tol=1e-5)
#     check_symmetric(hbb, raise_exception=True, tol=1e-5)
#     check_symmetric(hab, raise_exception=True, tol=1e-5)
#     print(kirch.shape, kbb.shape, kab.shape)
#     hess = aaGamma * haa + bbGamma * hbb + abGamma * hab
#
#     kirchoff = kirch * aaGamma + kbb * bbGamma + kab * abGamma
#     check_symmetric(hess, raise_warning=True, tol=1e-5)
#     print(hess.data)
#     hess.eliminate_zeros()
#
#     start = 3 * 0
#     stop = 3 * 30
#
#     fig, ax = plt.subplots()
#     mat = ax.matshow(haa[start:stop, start:stop].todense())
#     plt.colorbar(mat, ax=ax)
#     plt.show()
#     checkHessStrength(kirch, haa)
#
#     fig, ax = plt.subplots()
#
#     mat = ax.matshow(hbb[start:stop, start:stop].todense())
#     plt.colorbar(mat, ax=ax)
#     plt.show()
#     checkHessStrength(kbb, hbb)
#
#     fig, ax = plt.subplots()
#     mat = ax.matshow(hab[start:stop, start:stop].todense())
#     plt.colorbar(mat, ax=ax)
#     plt.show()
#     checkHessStrength(kab, hab)
#
#     fig, ax = plt.subplots()
#     mat = ax.matshow(hess[start:stop, start:stop].todense())
#     plt.colorbar(mat, ax=ax)
#     plt.show()
#     checkHessStrength(kirchoff, hess)
#
#     return kirchoff, hess
#
#
# def checkHessStrength(k, h):
#     """
#
#     :param k:
#     :param h:
#     """
#     hd = h.diagonal().reshape((-1, 3))
#     kd = k.diagonal()
#     hd = np.sum(hd, axis=-1)
#     print(np.min(hd))
#     print(hd)
#     print(kd)
#     print('total connections', np.sum(hd), np.sum(kd))
#     print('matrix sum', h.sum())
#
#
# # @nb.njit()
# def buildBetas(coords, calphas):
#     """
#
#     :param coords:
#     :param calphas:
#     :return:
#     """
#     na = calphas.numAtoms()
#     noBetas = []
#     aBetas = []
#     bvals = []
#     bcoords = np.zeros_like(coords)
#     count = 0
#     chid = 0
#     ch0 = calphas[0].getChid()
#     seg0 = calphas[0].getSegname()
#     nch = 0
#     for i, ca in enumerate(calphas.iterAtoms()):
#         if ca.getResname() == 'GLY':
#             # bcoords[i, :] = coords[i, :]
#             noBetas.append(i)
#             bvals.append(0)
#             # print('nobeta', i)
#             count += 1
#         elif count == 0 or i == na - 1:
#             # bcoords[i, :] = coords[i, :]
#             aBetas.append(i)
#             # noBetas.append(i)
#             # print('abeta', i)
#             # b = 1/np.linalg.norm(coords[i, :])
#             bvals.append(1)
#             count += 1
#         elif ca.getChid() == ch0 and ca.getSegname() == seg0 and calphas[i - 1].getChid() == ch0 and calphas[
#             i - 1].getSegname() == seg0 and calphas[i + 1].getChid() == ch0 and calphas[i + 1].getSegname() == seg0:
#             r = 2 * coords[i, :] - coords[i - 1, :] - coords[i + 1, :]
#             b = 1 / np.linalg.norm(r)
#             bvals.append(b)
#             bcoords[i, :] = coords[i, :] + 3.0 * r * b
#             count += 1
#         else:
#             bcoords[i, :] = coords[i, :]
#             aBetas.append(i)
#             bvals.append(1)
#             chid += 1
#             ch0 = ca.getChid()
#             seg0 = ca.getSegname()
#             count = 1
#             nch += 1
#     print(len(aBetas))
#
#     return bcoords, np.array(bvals), np.array(noBetas), np.array(aBetas)
#
#
# @nb.njit()
# def bbHessOnly(row, col, bcoords, nobetas, kGamma):
#     """
#
#     :param row:
#     :param col:
#     :param bcoords:
#     :param nobetas:
#     :param kGamma:
#     :return:
#     """
#     hData = np.zeros((row.shape[0], 3, 3))
#     dinds = np.nonzero(row == col)[0]
#     for k, (i, j) in enumerate(zip(row, col)):
#         if abs(i - j) <= 1 or np.any(nobetas == i) or np.any(nobetas == j):
#             continue
#
#         dvec = bcoords[j] - bcoords[i]
#         d2 = np.dot(dvec, dvec)
#         g = kGamma[k]
#         hblock = np.outer(dvec, dvec) * (g / d2)  # + (1-fanm)*np.identity(3)*g
#         hData[k, :, :] = hblock
#         hData[dinds[i]] += -hblock / 2
#         hData[dinds[j]] += -hblock / 2
#
#     return hData
#
#
# @nb.njit()
# def bbHess(row, col, bcoords, nobetas, bvals, abetas, kGamma, fanm):
#     """
#
#     :param row:
#     :param col:
#     :param bcoords:
#     :param nobetas:
#     :param bvals:
#     :param abetas:
#     :param kGamma:
#     :param fanm:
#     :return:
#     """
#     nrow = []
#     ncol = []
#     hData = []
#     krow = []
#     kcol = []
#     kData = []
#     n = bcoords.shape[0]
#
#     for k, (i, j) in enumerate(zip(row, col)):
#         if np.any(nobetas == j) or np.any(nobetas == i) or abs(i - j) <= 1:
#             continue
#         rvec = bcoords[i] - bcoords[j]
#         d2 = np.dot(rvec, rvec)
#         g = kGamma[k]
#         bi = bvals[i]
#         bj = bvals[j]
#         hblock = -(fanm * np.outer(rvec, rvec) + (1 - fanm) * np.identity(3)) * (
#                 g / d2) / 2  # + (1-fanm)*np.identity(3)*g
#         ijs = (3 * i, 3 * i + 3, 3 * i - 3, 3 * j, 3 * j + 3, 3 * j - 3)
#
#         if np.any(j == abetas):
#             cjs = (-1, 0, 0)
#         else:
#             cjs = (-(1 + 6 * bj), 3 * bj, 3 * bj)
#
#         if np.any(i == abetas):
#             cis = (1, 0, 0)
#         else:
#             cis = ((1 + 6 * bi), -3 * bi, -3 * bi)
#         cs = cis + cjs
#
#         for ci, ii in zip(cs, ijs):
#             if ci == 0:
#                 continue
#             for cj, jj in zip(cs, ijs):
#                 if cj == 0:
#                     continue
#                 krow.append(int(ii / 3))
#                 kcol.append(int(jj / 3))
#                 kData.append(-ci * cj * g)
#                 for l in range(3):
#                     for m in range(3):
#                         nrow.append(ii + l)
#                         ncol.append(jj + m)
#                         hData.append(ci * cj * hblock[l, m])
#
#     return nrow, ncol, hData, krow, kcol, kData
#
#
# @nb.njit()
# def abHessOnly(row, col, kGamma, bcoords, coords, nobetas, bvals, abetas, fanm):
#     """
#
#     :param row:
#     :param col:
#     :param kGamma:
#     :param bcoords:
#     :param coords:
#     :param nobetas:
#     :param bvals:
#     :param abetas:
#     :param fanm:
#     :return:
#     """
#     nrow = []
#     ncol = []
#     hData = []
#
#     krow = []
#     kcol = []
#     kData = []
#     for k, (i, j) in enumerate(zip(row, col)):
#         if np.any(nobetas == j) or abs(i - j) <= 1:
#             continue
#         rvec = coords[i] - bcoords[j]
#         d2 = np.dot(rvec, rvec)
#         g = kGamma[k]
#         b = bvals[j]
#         hblock = -(fanm * np.outer(rvec, rvec) + (1 - fanm) * np.identity(3)) * (g / d2)
#
#         ijs = (3 * i, 3 * j, 3 * j + 3, 3 * j - 3)
#
#         if np.any(j == abetas):
#             cjs = (-1., 0, 0)
#         else:
#             cjs = (-(1 + 6 * b), 3 * b, 3 * b)
#
#         cis = (1.,)
#
#         cs = cis + cjs
#
#         for ci, ii in zip(cs, ijs):
#             if ci == 0:
#                 continue
#             for cj, jj in zip(cs, ijs):
#                 if cj == 0.:
#                     continue
#                 krow.append(int(ii / 3))
#                 kcol.append(int(jj / 3))
#                 kData.append(-ci * cj * g)
#                 for l in range(3):
#                     for m in range(3):
#                         nrow.append(ii + l)
#                         ncol.append(jj + m)
#                         hData.append(ci * cj * hblock[l, m])
#
#     return nrow, ncol, hData, krow, kcol, kData
